! catcorr_aux.f90 Original author: Clive Page, cgp@leicester.ac.uk
! revised CGP 2012 March 9: to be compatible with catprep v0.3
! Revised CGP 2012 May 3: add LII_CORR and BII_CORR columns, v0.4
! Revised CGP 2012 May 21: add units to added columns
module catcorr_aux
use types
use errorHandling
use constants	       ! in ssclib, for Pi and TwoPi
use catcorr_minor	! defined in catprep
use nelder_mead
use sla			! defined in sla
implicit none
private
public catcorr_top
character(*), parameter :: VERSION = "0.4.3"
real(double), parameter :: HALFPI = 0.5_double * PI
real(double), parameter :: DTOR = PI/180.0_double, RTOD = 180.0_double/PI
real(double), parameter :: ASTOR = pi/(180.0_double * 3600.0_double), &
                           RTOAS = 1.0_double/ASTOR
integer, parameter  :: MAXSRC = 5000   ! Max number of X-ray detections
integer, parameter  :: MAXOPT = 100000 ! Max objects in any optical (or IR) catalogue
integer, parameter  :: MAXCAT = 3      ! Number of reference catalogues
real(double), parameter :: CATRADIUS = 14.0 * 60.0 * ASTOR  ! radius of field of view (rads)

type XlistType       ! structure for X-ray detections from one observation
                     ! derived from XMM FITS file of type OBSMLI
   character(10):: obs_id = ' '    ! observation identifier
   real(double) :: ra_pnt  = 0.0_double       ! RA of pointing axis (radians)
   real(double) :: dec_pnt = 0.0_double       ! Dec of Pointing axis (radians)
   real         :: year    = 2000.0           ! Years of obsvn e.g. 2009.3
   real(double) :: offset(3) = 0.0            ! RA/dec offsets, ROT_COR (RADIANS)
   real         :: offerr(3) = 0.0            ! errors in above (RADIANS)
   real(double) :: glat = 0.0_double          ! mean galactic latitude (RADIANS)
   integer      :: nsrc = 0                   ! Number of sources in following arrays
   logical      :: good(MAXSRC) = .false.     ! set .false. for poor quality detections
   integer      :: src_num(MAXSRC) = 0        ! Source number within this file
   real(double) :: ra(MAXSRC) = 0.0_double    ! RA from emldetect, J2000 (radians)
   real(double) :: dec(MAXSRC) = 0.0_double   ! Dec ditto , J2000 (radians)
   real         :: radec_err(MAXSRC) = 0.0    ! Standard error in position (radians)
   real         :: ep_extent(MAXSRC) = 0.0    ! Extent (pixels) =0.0 for point sources
   real         :: ep_tot(MAXSRC) = 0.0       ! EP_TOT total counts/sec
! Following values only set by OBSMLI_read if eposcorr/catcorr was run earlier
   logical      :: poscorok = .false.         ! From HDU2 header
   real         :: lik_hood  = 0.0            ! eposcorr likelihood
   integer      :: nmatches  = -1             ! NMATCHES in header
   real(double) :: ra_corr(MAXSRC) = 0.0_double  ! RA corrected by eposcorr (radians)
   real(double) :: dec_corr(MAXSRC) = 0.0_double ! Dec corrected by eposcorr (radians)
! Following values record catcorr results
   real(double) :: ra_cc(MAXSRC) = 0.0_double    ! Corrected ra
   real(double) :: dec_cc(MAXSRC) = 0.0_double   ! Corrected dec
   real(double) :: lii_corr(MAXSRC) = 0.0_double ! Galactic longitude ditto (radians)
   real(double) :: bii_corr(MAXSRC) = 0.0_double ! Galactic latitude ditto (radians)
   character(8) :: catused = '        '         ! Name of catalogue used for correction
end type XlistType
type(XlistType), save :: xlist
!
type catType          ! Storage for objects in the reference catalogues
   integer      :: nobj = 0             ! Number of catalogue objects in these arrays:
   character(10):: extname              ! Extension name
   real(double) :: cat_ra(MAXOPT)  = 0.0_double ! RA, J2000 (radians) 
   real(double) :: cat_dec(MAXOPT) = 0.0_double ! Dec, J2000 (radians)
   real         :: perr(MAXOPT) = 0.0   ! Std error on source postion (radians)
   real         :: bmag(MAXOPT) = 0.0   ! Blue magnitude
   real         :: rmag(MAXOPT) = 0.0   ! Red magnitude
end type catType
!
type(catType), save :: CATS(MAXCAT) ! storage for up to 3 catalogue extract,
! typically but not necessarily (1) USNO, (2) 2MASS, (3) SDSS
!
integer  :: NCAT = 0            ! Current catalogue used for fiting with MINIM
integer  :: SUMRESULT(-3:3) = 0 ! Counts how many times each catalogue used
integer  :: NOBSID = 0          ! Sequential count of observations for logging
integer  :: NMATCHES = 0        ! No of possible X-ray-opt matches
!
! Following arrays are written by cat_find_all_pairs and used by cross_match and
! Hold lists of feasible Opt/IR counterparts to each X-ray detn
! For the current reference catalogue only
integer, parameter :: MAXCP = 100    ! Max opt/IR counterparts per X-ray detn
integer  :: NCP(MAXSRC)              ! No of opt/IR counterparts per X-ray detn
integer  :: LISTCP(MAXCP, MAXSRC)    ! List of indexes of cat object 
!
CONTAINS
!--------------------------------------------------------------------------
subroutine catcorr_top(infile, catextract, mingood, minfit, maxOffset, status)
! Finds bets frame fit for one OBSMLI file using 3 catalogues
! If catextract FITS file is provided it uses that, else searches 3 cats itself
character(*), intent(in) :: infile     ! OBSMLI FITS file
character(*), intent(in) :: catextract ! Catalogue extract file (or blank)
integer, intent(in)      :: mingood    ! Minimum number of good detections
real, intent(in)         :: minfit     ! Minimum goodness of fit value to accept
real, intent(in)         :: maxOffset  ! Max frame offset to accept [arcsec]
integer, intent(inout)   :: status     ! Inherited status
!
integer, parameter  :: nop = 3, maxfn = 2000, iprint = -1, nloop = 6, iquad = 1
real(double), parameter :: stopcr = 1.0e-2_double, simp = 1.0e-5_double
real(double), parameter :: radius = 17.0_double * 60.0_double * ASTOR ! FOV radius, arcsecs
integer   :: nxgood, ifault, nxsrc, isrc, iresult, &
             catmatches(MAXCAT)
real(double)  :: bestfit, var(nop), param(nop), step(nop), off, cosdec, &
             goodness(MAXCAT), offset(nop,MAXCAT), offerr(nop,MAXCAT)
character(8), parameter :: catname(3) = (/ 'USNO-B1 ', '2MASS   ', 'SDSS-DR8' /)
!
if(status /= 0) RETURN
NOBSID = NOBSID + 1
! iresult codes:
! 1-3 best-fit catalogue is (1)=usno, (2)=2mass, (3)=sdss
! -1  too few good detections
! -2  goodness of fit below minfit
! -3  resulting offset too large
!
! Set up default values for cases where the fit fails
offset     = 0.0_double
offerr     = 0.0_double
goodness   = 0.0_double
catmatches = 0
call OBSMLI_read(infile, xlist, status)
call hist_angle(xlist, status)
nxsrc  = xlist%nsrc
nxgood = count(xlist%good(1:nxsrc))
if(nxgood < mingood) then
   iresult = -1
   NCAT    = 1   ! arbitrary choice so zeros get written to OBSMLI file
else
!
! get data from all 3 catalogues
!
call cat_read_extracts(catextract, CATS, status)
!
! For each catalogue with enough objects, match X-ray list and fit frame shift
!
   do NCAT = 1, MAXCAT
      if(CATS(NCAT)%nobj < 100) CYCLE
!
! To speed up cross-match, find for each xray source all catalogue sources
! nearby and store the information in module storage for cross_match
!
      call cat_find_all_pairs(xlist, maxOffset, CATS(NCAT))
!
! start with offsets zero so results not biased by eposcorr fitting
! initial stepsize values are RADIANS
!
      param = 0.0
      step   = (/0.1_double*ASTOR, 0.1_double*ASTOR, 0.05_double*dtor/)  
      call minim(param, step, nop, bestfit, maxfn, iprint, stopcr, nloop, &
              iquad, simp, var, cross_match, ifault)
      if(ifault == 0) then
         goodness(NCAT)   = abs(bestfit)
         offset(:,NCAT)   = param
         offerr(:,NCAT)   = sqrt(var)
         catmatches(NCAT) = NMATCHES
!        print *,'ncat=', ncat, 'offset=', param*rtod*3600.0
      end if
   end do
!
! Choose the catalogue which produced best goodness of fit
!
   NCAT = maxval(maxloc(goodness))  ! uses maxval() because maxloc returns array
   off  = sqrt((offset(1,NCAT)*cosdec)**2 + offset(2,NCAT)**2) * RTOAS
   if(goodness(NCAT) < minfit) then
      iresult = -2
      offerr(:,NCAT) = 0.0_double
   else if(off > maxOffset) then
      iresult = -3
   else                ! Everything ok, store results
      iresult   = NCAT
      cosdec    = cos(xlist%dec_pnt)
      xlist%catused = catname(ncat)
!
! save revised positions in data structure
!
      do isrc = 1,xlist%nsrc
         call rot_shift(xlist%ra_pnt, xlist%dec_pnt, offset(:,NCAT), &
            xlist%ra(isrc), xlist%dec(isrc), &
            xlist%ra_cc(isrc), xlist%dec_cc(isrc))
! And calculate galactic coordinates of these
         call sla_eqgal(xlist%ra_cc(isrc), xlist%dec_cc(isrc), &
                        xlist%lii_corr(isrc), xlist%bii_corr(isrc))
      end do
   end if
end if
!
! update the data structure and OBSMLI file 
!
xlist%offset = offset(:,NCAT)
xlist%offerr = offerr(:,NCAT)
xlist%lik_hood = goodness(NCAT)
xlist%nmatches = catmatches(NCAT)
xlist%poscorok = iresult > 0
call OBSMLI_update(infile, iresult, status)
SUMRESULT(iresult) = SUMRESULT(iresult) + 1
end subroutine catcorr_top
!-----------------------------------------------------------------------------
subroutine cat_find_all_pairs(xlist, maxOffset, cat)
! For each X-ray source, finds list of potential matches with optical or infra-red
! sources within CATRADIUS radians.  List of them saved in module storage
! for later use of routine cross_match
type(xlistType), intent(IN) :: xlist     ! list of X-ray detections
real, intent(in)            :: maxOffset ! Max offset to allow [arcsec]
type(catType), intent(IN)   :: cat       ! Selected single optical/ir catalogue
!
integer :: ix, icat
real(double) :: xra, xdec, offset, maxPairOffset
!
NCP    = 0
LISTCP = 0
! Store pairs with offsets up to maxOffset + 5 arcsecs, to cover the
! largest potential error circle.  Work through all "good" X-ray detections.
maxPairOffset = (maxOffset + 5.0) * ASTOR
do ix = 1, xlist%nsrc
   if(.not. xlist%good(ix)) CYCLE
   xra = xlist%ra(ix)
   xdec = xlist%dec(ix)
   do icat = 1, cat%nobj
   	offset = gcdist(xra, xdec, cat%cat_ra(icat), cat%cat_dec(icat))
   	if(offset < maxPairOffset) then
   	   if(NCP(ix) >= maxcp) then
             call warning('tooManyPairs', 'cat_find_all_pairs - too many pairs')
   	      EXIT
   	   end if
   	   NCP(ix) = NCP(ix) + 1
   	   LISTCP(NCP(ix), ix) = icat
      end if
   end do ! icat
end do ! ix
end subroutine cat_find_all_pairs
!-----------------------------------------------------------------------------
subroutine cat_read_extracts(filename, cat, status)
character(*), intent(in)   :: filename ! FITS file containing catalogue extracts
type(catType), intent(out) :: cat(:)   ! Catalogue data structure (up to 3 sets)
integer, intent(inout)     :: status   ! Inherited status
!
integer :: unit, icat, hdutype, nrows, kra, kdec, kperr, kbmag, krmag
logical :: anyf
character :: comment*60
!
if(status /= 0) RETURN
! Open the file READONLY
call ftgiou(unit, status)
call ftnopn(unit, filename, 0, status)
do icat = 1,3
! Move to HDU number (icat+1) containing a binary table, get number of rows
   call ftmahd(unit, icat+1, hdutype, status)
   if(status /= 0) then
! This HDU does not exist, set no of rows for the rest to zero, exit
      cat(icat:)%nobj = 0
      cat(icat:)%extname = ' '
      status = 0
      EXIT
   end if
   call ftgnrw(unit, nrows, status)
! Get the EXTNAME value of this HDU
   call ftgkys(unit, 'EXTNAME', cat(icat)%extname, comment, status)
! Read in the arrays, assumed to be in the order specified by cat_write_extracts
   cat(icat)%nobj = nrows
   if(nrows > 0) then
      call ftgcno(unit, .false., 'CAT_RA',        kra,  status)
      call ftgcno(unit, .false., 'CAT_DEC',       kdec, status)
      call ftgcno(unit, .false., 'CAT_RADEC_ERR', kperr, status)
      call ftgcno(unit, .false., 'BMAG',          kbmag, status)
      call ftgcno(unit, .false., 'RMAG',          krmag, status)
      call ftgcvd(unit, kra,   1, 1, nrows, 0.0d0, cat(icat)%cat_ra(1:nrows),   anyf, status)
      call ftgcvd(unit, kdec,  1, 1, nrows, 0.0d0, cat(icat)%cat_dec(1:nrows),  anyf, status)
      call ftgcve(unit, kperr, 1, 1, nrows, 0.0,   cat(icat)%perr(1:nrows), anyf, status)
      call ftgcve(unit, kbmag, 1, 1, nrows, 0.0,   cat(icat)%bmag(1:nrows), anyf, status)
      call ftgcve(unit, krmag, 1, 1, nrows, 0.0,   cat(icat)%rmag(1:nrows), anyf, status)
! Convert ra, dec, position error from degrees to radians
      cat(icat)%cat_ra(1:nrows)  = cat(icat)%cat_ra(1:nrows)  * DTOR
      cat(icat)%cat_dec(1:nrows) = cat(icat)%cat_dec(1:nrows) * DTOR
      cat(icat)%perr(1:nrows)    = cat(icat)%perr(1:nrows)    * DTOR
   end if
end do
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('cat_read_extracts', status)
end subroutine cat_read_extracts
!------------------------------------------------------------------------------
!+circle_around finds coordinates of NPTS points around a given point
subroutine circle_points(ra, dec, radius, rapt, decpt)
real(double), intent(in)  :: ra, dec, radius   ! (ra,dec) of centre, radius [radians]
real(double), intent(out) :: rapt(:), decpt(:) ! ra,dec of points (each size NPTS)
! in circle [radians] note wrapped so rapt() is close to ra even if -ve or
! over 2*pi radians in size, for convenience in finding map to use.
!
real(double) :: vz(3), vx(3), ctos(3,3), stoc(3,3), az, el, rapoint, decpoint
integer :: j, npts
!
! Get matrix to a new frame with this position at its centre, meridian arbitrary
!
npts = size(rapt)
call cona2v(ra, dec, vz)
call cona2v(ra, dec-0.5_double, vx)
call congen(vz, vx, ctos)
call conmin(ctos, stoc)
! Generate N points around a circle offset by radius from the pole
do j = 1,npts
   az = (twopi/npts) * j
   el = halfpi - radius
   call convrt(az, el, stoc, rapoint, decpoint)
! Wrap the RA value so that it is near the centre RA even if -ve or over 2pi
   rapt(j)  = mod(rapoint-ra + pi + twopi, twopi) + ra - pi
   decpt(j) = decpoint
end do
end subroutine circle_points
!-----------------------------------------------------------------------------
! cross_match: finds goodness of fit between X-ray detections and optical/IR
! counterparts, for current frame offset and reference catalogue NCAT
! Note: this routine is called for each iteration of MINIM
subroutine cross_match(param, result)
real(double), intent(in) :: param(:)  ! offsets in ra, dec, rotation (RADIANS)
real(double), intent(out) :: result   ! evaluated function at these offsets
!
real(double), parameter :: SYSERR = 0.35_double ! arcsecs ! systematic XMM error, arcsec
real(double), parameter :: SYSERRSQ = (SYSERR*ASTOR)**2  ! squared in radians
real(double), parameter :: max_fiterr = 2.0_double ! arcsecs
integer :: ix, icp, icat
real(double) :: offset, sigmasq, pmatch, pcoinc, net_prob, goodness, &
            best_prob, ra2, dec2
!
goodness = 0.0_double  ! Accumulates sum of individual net probabilities
NMATCHES = 0       ! Counts no of possible matches for export to FITS file
! Loop over all X-ray detections in the list
! But skip poor quality detections and those with zero potential counterparts
do ix = 1, xlist%nsrc
   if(NCP(ix) == 0 .or. .not. xlist%good(ix)) CYCLE
   call rot_shift(xlist%ra_pnt, xlist%dec_pnt, param, xlist%ra(ix), &
                  xlist%dec(ix), ra2, dec2)
! Loop over all potential opt/IR counterparts for this X-ray detection
   best_prob = 0.0_double
   do icp = 1, NCP(ix)
      icat = LISTCP(icp, ix)
!
! Apply the current parameter offset and rotation to the coordinates of this source
!
      offset  = gcdist(ra2, dec2, CATS(NCAT)%cat_ra(icat), CATS(NCAT)%cat_dec(icat))   ! radians
      sigmasq = xlist%radec_err(ix)**2 + CATS(NCAT)%perr(icat)**2 + SYSERRSQ  ! rads squared
      pmatch  = exp(-0.5 * offset**2 / sigmasq)
      pcoinc  = CATS(NCAT)%nobj * (offset/CATRADIUS)**2
      net_prob  = max(0.0_double, pmatch - pcoinc)
      best_prob = max(best_prob, net_prob)
   end do
   if(best_prob > 0.0_double) NMATCHES = NMATCHES + 1
   goodness = goodness + best_prob
end do
! Negate the goodness value as the fitting function MINIMISES the result 
result = -goodness
end subroutine cross_match
!-----------------------------------------------------------------------------
subroutine fits_status(subname, status)
! Reports on any non-zero FITS status value with message to standard output
character(*), intent(in) :: subname   ! Name of subroutine or function
integer, intent(in)      :: status    ! FITS status value to check
!
character :: msg*200
!
if(status == 0) RETURN
call ftgerr(status, msg)
write(errStr,'(a,i4,4a)') 'FITSIO error', status, ' in ', trim(subname), &
                     ' - ', trim(msg)
call error('fitsioError')
end subroutine fits_status
!-----------------------------------------------------------------------------
! hist_angle: Finds bad points from histogram of angles between pairs of points/
! Effect is that it resets some values in array xlist%good(:)
subroutine hist_angle(xlist, status)
type(XlistType), intent(inout) :: xlist 
integer, intent(inout)         :: status
!
integer, parameter :: nbins = 180, maxpt = 100000
integer :: j, k, n, ibin, nbad, &
   histogram(0:nbins-1), ipoint(0:nbins-1,maxpt), npoints(0:nbins-1), &
   badcount(MAXSRC)
real    :: dx, dy, theta, hmean, hlimit, ibinFloat
!
if(status /= 0) RETURN
histogram = 0
ipoint    = 0
npoints   = 0
!
! First produce a histogram of the angular distibution of position angles
!  between all distinct pairs of points (avoid double-counting pairs).
!  Note that the ipoint array stores the original list indexes of the points
!  forming each pair for each position angle bin for future track-back
!
do j = 1, xlist%nsrc
   do k = j+1, xlist%nsrc
      dy = xlist%dec(j) - xlist%dec(k)
      dx = (xlist%ra(j) - xlist%ra(k)) * cos(xlist%dec(j))
      theta = mod(twopi + atan2(dy,dx), pi) * rtod    ! theta range 0 to 180
      ibinFloat = nbins * theta / 180.0
      if (ibinFloat > nbins) then
        write(errStr, '(a,f11.5,a)') "catcorr_aux: ibin (", ibinFloat &
           , ") exceeds the upper limit.  Contact the code developer."
        call error('ibinTooLarge')
      else if (ibinFloat == nbins) then
        write(errStr, '(a,f11.5,a)') "catcorr_aux: ibin (", ibinFloat &
           , ") is larger than expected."
        call warning('ibinUnexpectedlyLarge')
      end if
      ! ibin = int(nbins * theta / 180.0)
      ibin = min(nbins-1, int(ibinFloat))	! To deal with floating point rounding cases.
      histogram(ibin) = histogram(ibin) + 1
      n = npoints(ibin)
      ipoint(ibin, n+1) = j
      ipoint(ibin, n+2) = k
      npoints(ibin) = n + 2
   end do
end do
!
! Choose suitable threshold for detecting a count excess
!
hmean = real(sum(histogram))/nbins
if(hmean < 20.0) then
   hlimit = poisson_level(hmean, 0.01/nbins)
else  ! Use Gaussian approximation
   hlimit = hmean + dinvnorm(0.01_double/nbins) * sqrt(hmean)
end if
!
! For each position angle at which there is an excess of pairs of points
! mark these by incrementing badcount(:)
!
badcount = 0
do ibin = 0, nbins-1
   if(histogram(ibin) > hlimit) then
! count all bad points
      do j = 1,npoints(ibin)
         k = ipoint(ibin,j)
         badcount(k) = badcount(k) + 1
      end do
   end if
end do
!
! for points with a high enough badpoint count, mark as not good in data
!
nbad = 0
do j = 1,xlist%nsrc
   if(badcount(j) >= 3) then
      xlist%good(j) = .false.
      nbad = nbad + 1
   end if
end do
end subroutine hist_angle
!-----------------------------------------------------------------------------
subroutine need_col(unit, colname, coltype, colunits, ncol, status)
! Finds a column by name, else creates new empty one of specified data type
integer, intent(in)      :: unit     ! Unit number of FITS file open to a table
character(*), intent(in) :: colname  ! Name of column
character(*), intent(in) :: coltype  ! Data type, e.g. '1D'
character(*), intent(in) :: colunits ! Physical units string, e.g. "degs"
integer, intent(out)     :: ncol     ! Column number if found/created
integer, intent(inout)   :: status   ! Inherited status
!
integer   :: ncolumns
character :: tunit*8
!
if(status /= 0) RETURN
call ftgcno(unit, .false., colname, ncol, status)
if(status /= 0) then
!
! the column does not exist, so add another column at the end
! 
   status = 0
   call ftgncl(unit, ncolumns, status)
   ncol = ncolumns + 1
   call fticol(unit, ncol, colname, coltype, status)
   write(tunit, '(a,i0)') 'TUNIT', ncol
   call ftukls(unit, tunit, colunits, "physical unit of the column", status)
end if
call fits_status('need_col', status)
end subroutine need_col
!-----------------------------------------------------------------------------
subroutine OBSMLI_read(OBSMLIfile, xlist, status)
! Reads selected columns of an OBSMLI file to the xlist structure.
character(*), intent(in)      :: OBSMLIfile      ! FITS file name of type OBSMLI
type(XlistType), intent(out)  :: xlist           ! selected contents of file
integer, intent(inout)        :: status          ! Inherited status
!
real, parameter :: radec_err_max = 3.0  ! Mark detection as bad if larger
integer :: nrows, nra, ndec, nradec_err, nep_extent, nep_tot, nsrc_num, &
           nra_corr, ndec_corr, hdutype, iyear, month, date, unit, nbii, j, &
           nbad
logical :: anyf
character :: comment*50, date_obs*30
real(double) :: ra_pnt, dec_pnt, bii(MAXSRC)
real     :: offset(3), offerr(3)
!
if(status /= 0) RETURN
xlist%nsrc   = 0
!
! Here open file READONLY - need to reopen later if we wish to update
!
call ftgiou(unit, status)
call ftnopn(unit, OBSMLIfile, 0, status) 
call fits_status('OBSMLI_read/ftnopn', status)
if(status /= 0) RETURN
!
! Get following scalars from FITS primary header
!
call ftgkys(unit, 'OBS_ID', xlist%obs_id, comment, status)
call ftgkyd(unit, 'RA_PNT', ra_pnt, comment, status)
call ftgkyd(unit, 'DEC_PNT', dec_pnt, comment, status)
call ftgkys(unit, 'DATE-OBS', date_obs, comment, status)
! Convert pointing axis coordinates from degrees to radians
xlist%ra_pnt = ra_pnt * dtor
xlist%dec_pnt = dec_pnt * dtor
call fits_status('OBSMLI_read/ftgkys', status)
if(status /= 0) go to 1000
!
! date of observation needed for proper-motion calculation, convert to real year
!
read(date_obs, '(i4,1x,i2,1x,i2)', iostat=status) iyear, month, date
if(status == 0) then
   xlist%year = 2000.0 + (mjd(iyear, month, date) - mjd(2000, 1, 1)) / 365.25
else
   call warning('invalidDateObs', 'error decoding DATE_OBS value')
   xlist%year = 2000.0
   date_obs = '?'
end if
!
! Move to Header Data Unit 2 and check if these headers exist from some
! earlier run of eposcorr or catcorr.
!
call ftmahd(unit, 2, hdutype, status)
call ftgkyl(unit, 'POSCOROK', xlist%poscorok, comment, status)
if(status == 0) then
   call ftgkye(unit, 'RAOFFSET', offset(1), comment, status)
   call ftgkye(unit, 'DEOFFSET', offset(2), comment, status)
   call ftgkye(unit, 'ROT_CORR', offset(3), comment, status)
   call ftgkye(unit, 'RAOFFERR', offerr(1), comment, status)
   call ftgkye(unit, 'DEOFFERR', offerr(2), comment, status)
   call ftgkye(unit, 'ROT_ERR ', offerr(3), comment, status)
   call ftgkye(unit, 'LIK_HOOD', xlist%lik_hood, comment, status)
   call ftgkyj(unit, 'NMATCHES', xlist%nmatches, comment, status)
!
! Convert all to radians and save in xlist data structure
!
   xlist%offset(1:2) = offset(1:2) * ASTOR
   xlist%offset(3)   = offset(3)   * dtor
   xlist%offerr(1:2) = offerr(1:2) * ASTOR
   xlist%offerr(3)   = offerr(3)   * dtor
else
   status = 0
   xlist%poscorok  = .false.
   xlist%offset    = 0.0
   xlist%offerr    = 0.0
   xlist%lik_hood  = 0.0
   xlist%nmatches  = -1      ! flag value meaning others are zero
end if
call fits_status('OBSMLI_read/ftgkye', status)
status = 0
!
! Get number of rows, and then column numbers of required columns
!
call ftgnrw(unit, nrows, status)
xlist%nsrc = nrows
call ftgcno(unit, .false., 'RA',        nra, status)
call ftgcno(unit, .false., 'DEC',       ndec, status)
call ftgcno(unit, .false., 'RADEC_ERR', nradec_err, status)
call ftgcno(unit, .false., 'EP_EXTENT', nep_extent, status)
call ftgcno(unit, .false., 'EP_TOT',    nep_tot, status)
call ftgcno(unit, .false., 'SRC_NUM',   nsrc_num, status)
call ftgcno(unit, .false., 'BII',       nbii, status)
!
! Read the arrays in
!
call ftgcvd(unit, nra,        1, 1, nrows, 0.0d0, xlist%ra, anyf, status)
call ftgcvd(unit, ndec,       1, 1, nrows, 0.0d0, xlist%dec, anyf, status)
call ftgcve(unit, nradec_err, 1, 1, nrows, 0.0,   xlist%radec_err, anyf, status)
call ftgcve(unit, nep_extent, 1, 1, nrows, 0.0,   xlist%ep_extent, anyf, status)
call ftgcve(unit, nep_tot,    1, 1, nrows, 0.0,   xlist%ep_tot, anyf, status)
call ftgcvj(unit, nsrc_num,   1, 1, nrows, 0,     xlist%src_num, anyf, status)
call ftgcvd(unit, nbii,       1, 1, nrows, 0.0_double, bii, anyf, status)
!
! Convert units to radians for RA, DEC, RADEC_ERR
!
xlist%ra(1:nrows)        = xlist%ra(1:nrows)        * dtor
xlist%dec(1:nrows)       = xlist%dec(1:nrows)       * dtor
xlist%radec_err(1:nrows) = xlist%radec_err(1:nrows) * ASTOR
!
! Find approx galactic latitude for the field by taking an average over detections
!
xlist%glat = dtor * SUM(bii(1:nrows)) / max(1,nrows)
!
! If POSCOROK exists, then get the actual corrected RA/DEC arrays
!
if(xlist%nmatches >= 0) then
   call ftgcno(unit, .false., 'RA_CORR',   nra_corr, status)
   call ftgcno(unit, .false., 'DEC_CORR',  ndec_corr, status)
   call ftgcvd(unit, nra_corr,   1, 1, nrows, 0.0_double, xlist%ra_corr, anyf, status)
   call ftgcvd(unit, ndec_corr,  1, 1, nrows, 0.0_double, xlist%dec_corr, anyf, status)
   xlist%ra_corr(1:nrows)   = xlist%ra_corr(1:nrows)   * dtor
   xlist%dec_corr(1:nrows)  = xlist%dec_corr(1:nrows)  * dtor
else
   xlist%ra_corr(1:nrows)  = xlist%ra(1:nrows)
   xlist%dec_corr(1:nrows) = xlist%dec(1:nrows)
end if
xlist%ra_cc(1:nrows)  = xlist%ra(1:nrows)
xlist%dec_cc(1:nrows) = xlist%dec(1:nrows)
call fits_status('OBSMLI_read/ftgcvd', status)
if(status /= 0) go to 1000
!
! Mark rows as bad if RADEC_ERR zero or too large, or source extended.
!
xlist%good(1:nrows) = .true.
nbad = 0
do j = 1,nrows
   if(xlist%radec_err(j)*RTOAS <= 0.001 .or. &
      xlist%radec_err(j)*RTOAS > radec_err_max .or. &
      xlist%ep_extent(j) > 0.1) then
         xlist%good(j) = .false.
         nbad = nbad + 1
   end if
end do
1000 continue
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('OBSMLI_read', status)
end subroutine OBSMLI_read
!-----------------------------------------------------------------------------
! OBSMLI_update: writes new positions to FITS file
subroutine OBSMLI_update(obsmlifile, iresult, status)
character(*), intent(in) :: obsmlifile ! to be opened and updated
integer, intent(in)      :: iresult    ! no of best catalogue (1-3) or -ve error
integer, intent(inout)   :: status     ! Inherited status
!
character(8), parameter :: catname(3) = (/ 'USNO-B1 ', '2MASS   ', 'SDSS-DR8' /)
integer :: unit, hdutype, nsrc, ncolra, ncoldec, idate(8), icat, ncolglong, &
           ncolglat
character :: datestring*20, comment*50
real      :: syserrcc
logical   :: first = .true.
!
if(status /= 0) return
icat = max(1,iresult)
nsrc = xlist%nsrc
if(xlist%poscorok) then
   syserrcc = 0.35
else
   syserrcc = 1.5
end if
! Open the FITS file for read/write
call ftgiou(unit, status)
call ftnopn(unit, OBSMLIfile, 1, status)
call fits_status('OBSMLI_update/ftopen', status)
!
! Insert history keyword with version number and date/time
!
call date_and_time(values=idate)
write(datestring, "(i4,2('-',i2.2),'T',i2.2,2('-',i2.2))") idate(1:3), idate(4:6)
call ftphis(unit, 'Modified by catcorr v'// version //' at '//datestring, &
            status)
!
! Move to 2nd HDU and check it is the right type i.e. binary table
!
call ftmahd(unit, 2, hdutype, status)
if(hdutype /= 2) then
   call error('obsmliNotTable', 'HDU2 not binary table as expected')
end if
!
! Check whether file has already been updated by e.g. eposcorr or catcorr
! by reading POSCOROK
!
call ftgkyl(unit, 'POSCOROK', xlist%poscorok, comment, status)
if(status /= 0) then
   status = 0
else if(first) then
   first = .false.
   call warning('alreadyCorrected', 'eposcorr/catcorr already run on this file')
end if
!
! Insert/update header keywords
!
call ftukyd(unit, 'RAOFFSET', xlist%offset(1)*RTOAS, 5, &
   'RA correction [arcsec]', status)
call ftukyd(unit, 'DEOFFSET', xlist%offset(2)*RTOAS, 5, &
   'DEC correction [arcsec]', status)
call ftukyd(unit, 'ROT_CORR', xlist%offset(3)*RTOD, 5, &
   'Rotation correction [degrees]', status)
call ftukyd(unit, 'RAOFFERR', xlist%offerr(1)*RTOAS, 5, &
   'Error in RA correction [arcsec]', status)
call ftukyd(unit, 'DEOFFERR', xlist%offerr(2)*RTOAS, 5, &
   'Error in Dec correction [arcsec]', status)
call ftukyd(unit, 'ROT_ERR ', xlist%offerr(3)*RTOD, 5, &
   'Error in rotation correction [degrees]', status)
call ftukye(unit, 'LIK_HOOD', xlist%lik_hood, 3, &
   'Goodness-of-fit statistic', status)
call ftukyj(unit, 'NMATCHES', xlist%nmatches, 'No of X-ray/ref matches', status)   
call ftukyl(unit, 'POSCOROK', xlist%poscorok, comment, status)
call ftukye(unit, 'SYSERRCC', syserrcc, 5, '[arcsec] systematic posn error', status)
call ftpkys(unit, 'CATUSED', xlist%catused, 'Catalogue used for correction', status)
!
! write updated positions converted to degrees from RA_CC/DEC_CC values
!
call need_col(unit, 'RA_CORR', '1D', 'deg', ncolra, status)
call ftpcld(unit, ncolra,  1, 1, nsrc, xlist%RA_CC(1:nsrc)*rtod, status)
!
call need_col(unit, 'DEC_CORR', '1D', 'deg', ncoldec, status)
! Error in following line fixed: ncolra ==> ncoldec
call ftpcld(unit, ncoldec, 1, 1, nsrc, xlist%DEC_CC(1:nsrc)*rtod, status)
! Ditto for galactics
call need_col(unit, 'LII_CORR', '1D', 'deg', ncolglong, status)
call ftpcld(unit, ncolglong, 1, 1, nsrc, xlist%lii_corr(1:nsrc)*rtod, status)
call need_col(unit, 'BII_CORR', '1D', 'deg', ncolglat, status)
call ftpcld(unit, ncolglat,  1, 1, nsrc, xlist%bii_corr(1:nsrc)*rtod, status)
!
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('OBSMLI_update', status)
end subroutine OBSMLI_update
!--
end module catcorr_aux
